// Generated by ResEdit 1.5.11
// Copyright (C) 2006-2012
// http://www.resedit.net

#include <windows.h>
/*
#include <commctrl.h>
#include <richedit.h>
*/
#include "w32_resource.h"




//
// Menu resources
//
LANGUAGE LANG_NEUTRAL, SUBLANG_NEUTRAL
IDR_MENU MENU
{
    POPUP "&File"
    {
        MENUITEM "&Consult...", IDM_CONSULT
        MENUITEM "C&hange Dir...", IDM_CHDIR
        MENUITEM "&Insert File Name...", IDM_FILE_NAME
        MENUITEM SEPARATOR
        MENUITEM "E&xit", IDM_EXIT
    }
    POPUP "&Edit"
    {
        MENUITEM "&Copy\tCtrl+C", IDM_COPY
        MENUITEM "&Paste\tCtrl+V", IDM_PASTE
        MENUITEM "&Select All\tAlt+A", IDM_SELECT_ALL
        MENUITEM "&Auto Copy on selection", IDM_COPY_ON_SEL, CHECKED
        MENUITEM SEPARATOR
        MENUITEM "Sa&ve Options", IDM_SAVE_OPTIONS
    }
    POPUP "&Terminal"
    {
        MENUITEM "&Beep on error", IDM_BEEP
        MENUITEM SEPARATOR
        MENUITEM "&Wrap mode (break long lines)\tAlt+W", IDM_WRAP
        MENUITEM SEPARATOR
        MENUITEM "&Line Buffering", IDM_BUFFERING, CHECKED
        MENUITEM "&Flush Output", IDM_FLUSH
        MENUITEM SEPARATOR
        MENUITEM "&Show Text Console", IDM_SHOW_CONSOLE
        MENUITEM SEPARATOR
        MENUITEM "&Change Font...", IDM_FONT
    }
    POPUP "Prolog"
    {
        MENUITEM "&Interrupt (send ^C)\tCtrl+Alt+C", IDM_INTERRUPT
        MENUITEM "&Set Prolog memory sizes", IDM_STACK_SIZES
    }
    POPUP "&Help"
    {
        MENUITEM "&Manual\tF1", IDM_MANUAL
        MENUITEM "&Index Search\tShift+F1", IDM_INDEX
        MENUITEM SEPARATOR
        MENUITEM "&GNU Prolog web page", IDM_WEB
        MENUITEM SEPARATOR
        MENUITEM "&About", IDM_ABOUT
    }
}



//
// Dialog resources
//
LANGUAGE LANG_NEUTRAL, SUBLANG_NEUTRAL
IDD_STACK_SIZES DIALOG 20, 20, 444, 222
STYLE DS_MODALFRAME | DS_SHELLFONT | WS_CAPTION | WS_POPUP | WS_SYSMENU
CAPTION "Memory sizes setting"
FONT 9, "Ms Shell Dlg"
{
    DEFPUSHBUTTON   "&OK", IDOK, 306, 197, 57, 14
    PUSHBUTTON      "&Cancel", IDCANCEL, 372, 197, 57, 14
    EDITTEXT        IDC_REG_SZ0, 306, 60, 57, 12, ES_RIGHT | ES_AUTOHSCROLL | ES_NUMBER, WS_EX_STATICEDGE
    RTEXT           "", IDC_STACK_DESC0, 18, 59, 78, 12, SS_RIGHT
    RTEXT           "", IDC_DEF_SZ0, 111, 59, 57, 12, WS_BORDER | SS_RIGHT
    RTEXT           "", IDC_ENV_SZ0, 241, 60, 57, 12, WS_BORDER | SS_RIGHT
    LTEXT           "", IDC_ENV_VAR_NAME0, 176, 60, 57, 12, WS_BORDER | SS_LEFT
    CTEXT           "Default size", IDC_STATIC, 110, 43, 57, 10, SS_CENTER
    CTEXT           "Var/Key name", IDC_STATIC, 177, 43, 56, 10, SS_CENTER
    CTEXT           "Env. var. value", IDC_STATIC, 244, 43, 49, 8, SS_CENTER
    CTEXT           "Registry value", IDC_STATIC, 312, 43, 46, 8, SS_CENTER
    LTEXT           "You can adjust the size of GNU Prolog memory using environment variables and/or registry keys. Here you can modify registry values.\nFor stacks, sizes are in kilobytes (Kb). For the atom table, it is the number of atoms.", IDC_STATIC, 8, 9, 419, 16, SS_LEFT
    RTEXT           "", IDC_STACK_DESC1, 18, 86, 78, 12, SS_RIGHT
    RTEXT           "", IDC_DEF_SZ1, 111, 86, 57, 12, WS_BORDER | SS_RIGHT
    LTEXT           "", IDC_ENV_VAR_NAME1, 176, 86, 57, 12, WS_BORDER | SS_LEFT
    RTEXT           "", IDC_ENV_SZ1, 241, 86, 57, 12, WS_BORDER | SS_RIGHT
    EDITTEXT        IDC_REG_SZ1, 306, 86, 57, 12, ES_RIGHT | ES_AUTOHSCROLL | ES_NUMBER, WS_EX_STATICEDGE
    RTEXT           "", IDC_STACK_DESC2, 18, 113, 78, 12, SS_RIGHT
    RTEXT           "", IDC_DEF_SZ2, 111, 113, 57, 12, WS_BORDER | SS_RIGHT
    LTEXT           "", IDC_ENV_VAR_NAME2, 176, 113, 57, 12, WS_BORDER | SS_LEFT
    RTEXT           "", IDC_ENV_SZ2, 241, 113, 57, 12, WS_BORDER | SS_RIGHT
    EDITTEXT        IDC_REG_SZ2, 306, 113, 57, 12, ES_RIGHT | ES_AUTOHSCROLL | ES_NUMBER, WS_EX_STATICEDGE
    RTEXT           "", IDC_STACK_DESC3, 18, 141, 78, 12, SS_RIGHT
    RTEXT           "", IDC_DEF_SZ3, 111, 141, 57, 12, WS_BORDER | SS_RIGHT
    LTEXT           "", IDC_ENV_VAR_NAME3, 176, 141, 57, 12, WS_BORDER | SS_LEFT
    RTEXT           "", IDC_ENV_SZ3, 241, 141, 57, 12, WS_BORDER | SS_RIGHT
    EDITTEXT        IDC_REG_SZ3, 306, 141, 57, 12, ES_RIGHT | ES_AUTOHSCROLL | ES_NUMBER, WS_EX_STATICEDGE
    RTEXT           "", IDC_CUR_SZ0, 372, 59, 57, 12, WS_BORDER | SS_RIGHT
    CTEXT           "Current size", IDC_STATIC, 383, 43, 38, 8, SS_CENTER
    RTEXT           "", IDC_CUR_SZ1, 372, 86, 57, 12, WS_BORDER | SS_RIGHT
    RTEXT           "", IDC_CUR_SZ2, 372, 113, 57, 12, WS_BORDER | SS_RIGHT
    RTEXT           "", IDC_CUR_SZ3, 372, 141, 57, 12, WS_BORDER | SS_RIGHT
    LTEXT           "", IDC_FIXED_SIZES, 9, 198, 290, 13, SS_LEFT | SS_NOPREFIX
    RTEXT           "", IDC_STACK_DESC4, 18, 170, 78, 12, SS_RIGHT
    RTEXT           "", IDC_DEF_SZ4, 111, 170, 57, 12, WS_BORDER | SS_RIGHT
    LTEXT           "", IDC_ENV_VAR_NAME4, 176, 170, 57, 12, WS_BORDER | SS_LEFT
    RTEXT           "", IDC_ENV_SZ4, 241, 170, 57, 12, WS_BORDER | SS_RIGHT
    EDITTEXT        IDC_REG_SZ4, 306, 170, 57, 12, ES_RIGHT | ES_AUTOHSCROLL | ES_NUMBER, WS_EX_STATICEDGE
    RTEXT           "", IDC_CUR_SZ4, 372, 170, 57, 12, WS_BORDER | SS_RIGHT
}



//
// Accelerator resources
//
LANGUAGE LANG_NEUTRAL, SUBLANG_NEUTRAL
IDR_ACCEL ACCELERATORS
{
    "C",            IDM_INTERRUPT, VIRTKEY, ALT
    "C",            IDM_INTERRUPT, VIRTKEY, ALT, CONTROL
    "C",            IDM_COPY, VIRTKEY, CONTROL
    "V",            IDM_PASTE, VIRTKEY, CONTROL
    "A",            IDM_SELECT_ALL, VIRTKEY, ALT
    "W",            IDM_WRAP, VIRTKEY, ALT
    VK_PAUSE,      IDM_INTERRUPT, VIRTKEY, CONTROL
    VK_F1,         IDM_MANUAL, VIRTKEY
    VK_F1,         IDM_INDEX, VIRTKEY, SHIFT
}



//
// Icon resources
//
LANGUAGE LANG_NEUTRAL, SUBLANG_NEUTRAL
IDI_ICON           ICON           "../../gprolog.ico"
