
:-	include(read_file).
:-	include(bip_list).
:-	include(syn_sugar).
:-	include(internal).
:-	include(code_gen).
:-	include(reg_alloc).
:-	include(inst_codif).
:-	include(first_arg).
:-	include(indexing).
:-	include(wam_emit).
:-	include(pl2wam).
